/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"


/* wrappers for fftw init and cleanup */
void PX(init) (void){
  XM(init)();
}

void PX(cleanup) (void){
  XM(cleanup)();
}


/* functions to execute and destroy PX(plan) */
void PX(execute_timed)(
    PX(plan) fftplan, double *time)
{ /* give back timers of the main parts */
  int p;

  if(fftplan==NULL){
    fprintf(stderr, "!!! Error: FFT3D_Plan == NULL !!!\n");
    return;
  }

  for(p=0; p<3; p++){
    time[2*p] = -MPI_Wtime();
    if(fftplan->local_fft[p] != NULL)
      PX(serfft_trunc_execute)(fftplan->local_fft[p]);
    time[2*p] += MPI_Wtime();

    time[2*p+1] = -MPI_Wtime();
    if(fftplan->global_remap[p] != NULL)
      X(execute)(fftplan->global_remap[p]);
    time[2*p+1] += MPI_Wtime();
  }
} /* end PX(execute_timed) */


void PX(execute) (PX(plan) fftplan){
  int p;

  if(fftplan==NULL){
    fprintf(stderr, "!!! Error: FFT3D Plan == NULL !!!\n");
    return;
  }

  for(p=0; p<3; p++){
    if(fftplan->local_fft[p] != NULL)
      PX(serfft_trunc_execute)(fftplan->local_fft[p]);
    if(fftplan->global_remap[p] != NULL)
      X(execute)(fftplan->global_remap[p]);
  }
} /* end PX(forw_execute) */


void PX(destroy_plan) (PX(plan) fftplan){
  int p;

  if(fftplan==NULL){
    fprintf(stderr, "!!! Error: Can not finalize PFFT Plan == NULL !!!\n");
    return;
  }

  for(p=0; p<3; p++){
    if(fftplan->local_fft[p] != NULL)
      PX(serfft_trunc_destroy_plan)(fftplan->local_fft[p]);
    if(fftplan->global_remap[p] != NULL)
      X(destroy_plan)(fftplan->global_remap[p]);
  }
  
  PX(destroy_gcplan)(fftplan->gcplan);
  
  PX(free)(fftplan);
  
  fftplan = NULL;
} /* end PX(destroy_plan) */


void PX(init_input_3d)(
    const INT *n, const INT *local_n, const INT *local_n_start, MPI_Comm comm,
    C *data
    )
{ /* initialize FFT input with random numbers */
  PX(init_many_input_3d)(
      n, n, local_n, local_n_start, comm,
      data);
}

void PX(check_output_3d)(
    C *data, const INT *n,
    const INT *local_n, const INT *local_n_start,
    MPI_Comm comm, int iterations, int is_scaled
    )
{ /* check results after one forward and backward FFT, only works for intial data layout */
  PX(check_many_output_3d)(
      data, n, n, local_n, local_n_start, comm, iterations, is_scaled);
}


INT PX(local_size_3d)(
    const INT *n,
    MPI_Comm comm_cart_2d, unsigned pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{
  INT alloc_local_data_in, howmany = 1; /* howmany>1 only for d>3 */

  return PX(local_size_3d_many)(
      n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      PFFT_NO_GCELLS, PFFT_NO_GCELLS, comm_cart_2d, pfft_flags,
      local_ni, local_i_start, local_no, local_o_start, &alloc_local_data_in);
}


PX(plan) PX(plan_dft_3d)(
    const INT *n, C *data_in, C *data_out,
    MPI_Comm comm_cart_2d,
    int sign, unsigned pfft_flags, unsigned fftw_flags
    )
{ /* !!! with flag FFTW_MEASURE planning will overwrite users data !!! */
  INT howmany = 1; /* howmany>1 only for d>3 */

  return PX(plan_many_dft_3d)(
      n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, PFFT_NO_GCELLS, PFFT_NO_GCELLS,
      data_in, data_out, comm_cart_2d, sign, pfft_flags, fftw_flags);
}


int PX(check)(
    const INT *n, MPI_Comm comm_cart_2d, unsigned pfft_flags, int verbose
    )
{
  return PX(check_many)(n, n, n, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart_2d, pfft_flags, verbose);
}


void PX(exchange_ghostcells)(
    PX(plan) ths
    )
{
  PX(exchange_ghostcells_internal)(ths->gcplan);
}


void PX(reduce_ghostcells)(
    PX(plan) ths
    )
{
  PX(reduce_ghostcells_internal)(ths->gcplan);
}



static void print_complex_array(
     double *data, INT *n, INT *start, char *name
     )
{
  INT k0, k1, k2, l;
  
  for(k0 = 0, l = 0; k0 < n[0]; k0++){
    printf("%s(%td,%td:%td,%td:%td):", name, k0+start[0]+1, start[1]+1,
           start[1]+n[1], start[2]+1, start[2]+n[2]);
    for(k1 = 0; k1 < n[1]; k1++){
      for(k2 = 0; k2 < n[2]; k2++, l++){
        if(k2%4 == 0)
          printf("\n%4td.", l/4);
        printf(" %.2e+%.2ei,", data[2*l], data[2*l+1]);
      }
    }
    printf("\n");
  }
}


void PX(apr_complex_3d)(
     fftw_complex *data, INT *local_n, INT *local_n_start,
     char *name, MPI_Comm comm
     )
{
  int num_procs, proc_rank;
  
  if( PX(prod_ptrdiff_t)(3, local_n) < 1)
    return;

  MPI_Comm_size(comm, &num_procs);
  MPI_Comm_rank(comm, &proc_rank);

  fflush(stdout);
  for(int t=0; t<num_procs; t++){
    if(proc_rank == t){
      printf("\nRank %d: ", proc_rank);
      print_complex_array((double*) data, local_n, local_n_start, name);
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PX(apr_complex_permuted_3d)(
     fftw_complex *data, INT *local_n, INT *local_n_start,
     int perm0, int perm1, int perm2,
     char *name, MPI_Comm comm
     )
{
  int perm[3];
  INT local_n_start_perm[3], local_n_perm[3];
  
  perm[0] = perm0; perm[1] = perm1; perm[2] = perm2;
  
  for(int t=0; t<3; t++){
    local_n_perm[t] = local_n[perm[t]];
    local_n_start_perm[t] = local_n_start[perm[t]];
  }
  
  PX(apr_complex_3d)(data, local_n_perm, local_n_start_perm, name, comm);
}

