/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

static int check_fft_1d(
    unsigned fft_flag, int num_fft_1d, int verbose, PX(permutation) *perm);
static int check_fft1d_num_transpose_flags(
    unsigned fft_flag, int num_fft, int verbose);
static int check_transpose(
    const INT* n, const INT* block, MPI_Comm comm_cart_1d,
    int first_dims, unsigned remap_flags, int num_remap, int verbose,
    PX(permutation) *perm);


int PX(check_cart_2d_internal)(
    MPI_Comm comm_cart_2d, int verbose
    )
{
  int proc_rank, errorcode = 0;

  MPI_Comm_rank(comm_cart_2d, &proc_rank);
  if( ! PX(is_cart_procmesh_2d)(comm_cart_2d) ){
    if( verbose && !proc_rank)
      fprintf(stderr, "Error in PFFT: Communicator has no 2d Cartesian topology\n");
    errorcode = 1;
  }

  return errorcode;
}

int PX(check_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    const INT *block_col, const INT *block_row, MPI_Comm comm_col, MPI_Comm comm_row,
    int verbose
    )
{
  int coord_col, coord_row, num_comm;
  int retval=0;
  const INT *block;
  PX(permutation) perm_out;
  PX(truncation) trc_out;
  MPI_Comm comm;

  PX(get_mpi_cart_coord)(comm_col, &coord_col);
  PX(get_mpi_cart_coord)(comm_row, &coord_row);
  verbose *= !coord_col * !coord_row;
  
  PX(copy_permutation)(perm, &perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  
  /* avoid senseless first_dims parameter */
  for(int p=0; p<3; p++){
    if( (prf->first_dims[p] < 0) || (prf->first_dims[p] > 2) ){
      if( verbose )
        fprintf(stderr, "ERROR in PFFT: prf->first_dims[%d] must be equal to 0, 1 or 2\n", p);
      return 1;
    }
  }

  for(int p=0; p<3; p++){
    /* check if every 1d FFT is calculated */
    PX(mark_performed_fft)(&perm_out, &trc_out);

    /* check 1d FFT and permute */
    retval = check_fft_1d(prf->fft_flag[p], p+1, verbose, &perm_out);
    if(retval==1) return 1;
    
    /* check remap and permute */
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;

    retval = check_transpose(n, block, comm, prf->first_dims[p],
        prf->remap_flag[p], p+1, verbose, &perm_out);
    if(retval==1) return 1;
  }
  
  for(int p=0; p<3; p++){
    if(trc_out.is_transformed[p] != 1){
      retval=1;
      if(verbose)
        fprintf(stderr, "Error in PFFT: %d FFTs along n[%d] !!!\n", trc_out.is_transformed[p], p);
    }
  }
  
  return retval;
}


static int check_fft_1d(
    unsigned fft_flag, int num_fft_1d, int verbose, PX(permutation) *perm
    )
{
  int dim_fft, retval = 0;
  
  dim_fft = PX(get_local_dim)(perm);
  
  retval = check_fft1d_num_transpose_flags(fft_flag, num_fft_1d, verbose);
  
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_201) && (dim_fft != 2) ){
    if(retval == 0) retval = 2;
    if( verbose )
      fprintf(stderr, "!!! Warning in PFFT: %d. FFT, TRANSPOSE_201 works in-place only with FFT on last dim !!!\n", num_fft_1d);
  }
  
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_120) && (dim_fft != 0) ){
    if(retval == 0) retval = 2;
    if( verbose )
      fprintf(stderr, "!!! Warning in PFFT: %d. FFT, TRANSPOSE_120 works in-place only with FFT on first dim !!!\n", num_fft_1d);
  }
  
  PX(permute_fft_1d)(fft_flag, perm);
  
  return retval;
}


static int check_fft1d_num_transpose_flags(
    unsigned fft_flag, int num_fft, int verbose)
{
  int retval = 0, count = 0;

  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_102) ) count++;
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_021) ) count++;
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_201) ) count++;
  if( PX(flag_active)(fft_flag, PFFT_TRANSPOSE_120) ) count++;

  if(count > 1){
    retval = 1;
    if( verbose )
      fprintf(stderr, "!!! ERROR in PFFT: %d. FFT gets more than one transpose flag !!!\n", num_fft);
  }

  return retval;
}

static int check_transpose(
    const INT* n, const INT* block, MPI_Comm comm_cart_1d,
    int first_dims, unsigned remap_flags, int num_remap, int verbose,
    PX(permutation) *perm
    )
{
  int num_procs, num_comm, local_dim, needed_local_dim, retval=0;
  INT N[2], blk[2];

  if( (first_dims != 1) && (first_dims != 2) )
    return 0;

  MPI_Comm_size(comm_cart_1d, &num_procs);
  num_comm = PX(remap_num_comm)(first_dims, remap_flags, perm);

  local_dim = PX(get_local_dim)(perm);
  needed_local_dim = PX(flag_active)(remap_flags, FFTW_MPI_TRANSPOSED_IN) ? 0 : first_dims;
  if( local_dim != needed_local_dim ){
    retval = 1;
    if(verbose)
      fprintf(stderr, "!!! Error in PFFT: %d. FFT returns local dim %d, but %d. remap needs local dim %d !!!\n",
              num_remap, local_dim, num_remap, needed_local_dim);
  }


  if( PX(flag_active)(remap_flags, FFTW_MPI_TRANSPOSED_IN) ){
    PX(permute_transposed_in_or_out)(first_dims, perm);
    first_dims = 3 - first_dims; /* set first_dims to opposite value */
  }

  /* set N and blk for FFTW */
  N[0] = n[perm->n[0]];
  N[1] = n[perm->n[first_dims]];
  blk[0] = block[perm->n[0]];
  blk[1] = block[perm->n[first_dims]];


  if( (num_procs > PX(num_blocks)(N[0], blk[0])) && (num_procs > PX(num_blocks)(N[1], blk[1])) ){
    if(retval==0) retval = 2;
    if(verbose){
      fprintf(stderr, "!!! Warning in PFFT: Before AND after %d. remap some processors have no data !!!\n", num_remap);
      fprintf(stderr, "!!! Warning in PFFT: This will cause FFTW to hang up for out of place transforms!!!\n");
    }
  }

  if( num_procs>N[0] || num_procs>N[1] ){
    if(retval==0) retval = 2;
    if(verbose){
      printf("!!! Warning in PFFT: inefficient processor mesh for %d. remap !!!\n", num_remap);
      printf("!!! Warning in PFFT: please check np%d <= n[%d] and np%d <= n[%d] !!!\n",
             num_comm, perm->n[0], num_comm, perm->n[first_dims]);
    }
  }

  PX(permute_std_transpose_2d)(first_dims, perm);
  
  if( PX(flag_not_active)(remap_flags, FFTW_MPI_TRANSPOSED_OUT) )
    PX(permute_transposed_in_or_out)(first_dims, perm);

  return retval;
}

