/*
 * Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

INT PX(local_size_dft_3d_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    MPI_Comm comm_col, MPI_Comm comm_row,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start
    )
{ /* get local data size of given profile */
  int num_comm;
  INT alloc_local_transp[3], alloc_local_fft[3], alloc_local_max, local_n[3], n_trunc[3];
  const INT *block;
  PX(permutation) perm_out;
  PX(truncation) trc_out;
  MPI_Comm comm;
  
  PX(copy_permutation)(perm, &perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  for(int p=0; p<3; p++){
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    alloc_local_fft[p] = PX(local_size_fft_1d)(
        local_n, tuple_size, prf->fft_flag[p], &trc_out, &perm_out);
    
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;
    PX(get_current_global_size)(&trc_out, n_trunc);

    alloc_local_transp[p] = PX(local_size_transpose_2d)(
        n_trunc, local_no, tuple_size, block, prf->first_dims[p], comm, prf->remap_flag[p], &perm_out);
  }

  /* get initial and final data offsets */
  PX(decompose_3d)(trc->ni, perm, block_col, block_row, comm_col, comm_row,
      local_ni, local_i_start);
  PX(decompose_3d)(trc->no, &perm_out, block_col, block_row, comm_col, comm_row,
      local_no, local_o_start);

  /* return maximum needed memory */
  alloc_local_max = 0;
  for(int t=0; t<3; t++){
    alloc_local_max = MAX(alloc_local_max, alloc_local_transp[t]);
    alloc_local_max = MAX(alloc_local_max, alloc_local_fft[t]);
  }

  return alloc_local_max;
}


PX(plan) PX(plan_dft_3d_internal)(
    const INT *n, const PX(profile) *prf,
    const PX(truncation) *trc, const PX(permutation) *perm,
    INT tuple_size, const INT *block_col, const INT *block_row,
    C *data_in, C *data_out, MPI_Comm comm_col, MPI_Comm comm_row,
    int sign, unsigned fftw_planning_flags
    )
{ /* !!! with flag FFTW_MEASURE planning will overwrite users data !!! */
  int num_comm;
  INT local_ni[3], local_n[3], local_no[3], n_trunc[3];
  const INT *block;
  PX(permutation) perm_out;
  PX(truncation) trc_out;
  pfft_plan_internal* fftplan;
  MPI_Comm comm;

  fftplan = (pfft_plan_internal*) PX(malloc)(sizeof(pfft_plan_internal), comm_row);
  
  PX(copy_permutation)(perm, &perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  
  /* split planing of 1dFFTs and remaps to avoid long planing
  times caused by forgotten wisdom in front of remap planing */
  for(int p=0; p<3; p++){
    /* get data layout */
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    /* plan 1D-FFT */
    fftplan->local_fft[p] = PX(plan_dft_trunc_1d)(
        local_ni, local_n, local_no, tuple_size, data_in, data_out,
        sign, prf->fft_flag[p], fftw_planning_flags,
        &trc_out, &perm_out);

    /* simulate remap */
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;

    PX(permute_transpose_2d)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
  }
  
  /* reinitialize data distribution and permutation */
  PX(copy_permutation)(perm, &perm_out);
  PX(init_truncation)(trc->ni, trc->no, &trc_out);
  for(int p=0; p<3; p++){
    /* get data layout */
    PX(decompose_fft_trunc_3d)(
        n, &trc_out, &perm_out, block_col, block_row, comm_col, comm_row,
        local_ni, local_n, local_no);

    /* simulate 1D-FFT */
    PX(permute_fft_1d)(prf->fft_flag[p], &perm_out);
    PX(mark_performed_fft)(&perm_out, &trc_out);
    
    /* plan remap */
    X(forget_wisdom)();
    num_comm = PX(remap_num_comm)(prf->first_dims[p], prf->remap_flag[p], &perm_out);
    block = (num_comm == 0) ? block_col : block_row;
    comm  = (num_comm == 0) ? comm_col  : comm_row;
    PX(get_current_global_size)(&trc_out, n_trunc);
    
    fftplan->global_remap[p] =  PX(plan_transpose_2d)(
        n_trunc, local_no, tuple_size, block, data_out, data_in, prf->first_dims[p], comm,
        fftw_planning_flags| prf->remap_flag[p], &perm_out);

    /* FIXME: ??? possible FFTW BUG ????:
     * something goes wrong in FFTWs wisdom management for parallel transposition,
     * sometimes next planing freezes, delete wisdom to avoid this */
    X(forget_wisdom)();
  }

  if( PX(is_illegal_plan)(fftplan) ){
      PX(destroy_plan)(fftplan);
      fftplan = NULL;
  }

  return fftplan;
}


void PX(permute_fft_3d)(
    const PX(profile) *prf, PX(permutation) *perm
    )
{
  for(int p=0; p<3; p++){
    PX(permute_fft_1d)(prf->fft_flag[p], perm);
    PX(permute_transpose_2d)(prf->first_dims[p], prf->remap_flag[p], perm);
  }
}


