/*
 * Copyright (c) 2003, 2007-8 Matteo Frigo
 * Copyright (c) 2003, 2007-8 Massachusetts Institute of Technology
 * Copyright (c) 2010 Michael Pippig
 *
 * The following statement of license applies *only* to this header file,
 * and *not* to the other files distributed with FFTW or derived therefrom:
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


#ifndef __PFFT_H__
#define __PFFT_H__

#include "fftw3.h"
#include <mpi.h>

#ifdef __cplusplus
  extern "C"
  {
#endif /* __cplusplus */

#define PFFT_EXTERN extern
#define PFFT_CONCAT(prefix, name) prefix ## name


/*
  huge second-order macro that defines prototypes for all API
  functions.  We expand this macro for each supported precision
 
  PX: PFFT name-mangling macro (parallel)
  X : FFTW name-mangling macro (serial)
  R : real data type
  C : complex data type
*/
#define PFFT_DEFINE_API(PX, X, R, C)                                                    \
                                                                                        \
  typedef struct PX(plan_s) *PX(plan);                                                  \
  typedef struct{                                                                       \
    ptrdiff_t ni[3];                                                                    \
    ptrdiff_t no[3];                                                                    \
    int is_transformed[3];                                                              \
  } PX(truncation);                                                                     \
  typedef struct{                                                                       \
    int n[3];                                                                           \
    int np[3];                                                                          \
  } PX(permutation);                                                                    \
  typedef struct {                                                                      \
    unsigned fft_flag[3];                                                               \
    int first_dims[3];                                                                  \
    unsigned remap_flag[3];                                                             \
  } PX(profile);                                                                        \
                                                                                        \
  PFFT_EXTERN void PX(init)(void);						        \
  PFFT_EXTERN void PX(cleanup)(void);                                                   \
                                                                                        \
  PFFT_EXTERN void PX(execute_timed)(PX(plan) fftplan, R *time); 			\
  PFFT_EXTERN void PX(execute)(PX(plan) fftplan);					\
  PFFT_EXTERN void PX(destroy_plan)(PX(plan) fftplan);                                  \
                                                                                        \
  PFFT_EXTERN int PX(create_procmesh_2d)(                                               \
      MPI_Comm comm, int np0, int np1,                                                  \
      MPI_Comm *comm_cart_2d);                                                          \
                                                                                        \
  PFFT_EXTERN void PX(init_input_3d)(                                                   \
      const ptrdiff_t *n, const ptrdiff_t *local_n, const ptrdiff_t *local_n_start,     \
      MPI_Comm comm, C *data);                                                          \
  PFFT_EXTERN void PX(init_many_input_3d)(                                              \
      const ptrdiff_t *n, const ptrdiff_t *ni,                                          \
      const ptrdiff_t *local_n, const ptrdiff_t *local_n_start, MPI_Comm comm,          \
      C *data);                                                                         \
  PFFT_EXTERN void PX(check_output_3d)(                                                 \
      C *data, const ptrdiff_t *n,                                                      \
      const ptrdiff_t *local_n, const ptrdiff_t *local_n_start,                         \
      MPI_Comm comm, int iterations, int iscale);                                       \
  PFFT_EXTERN void PX(check_many_output_3d)(                                            \
      C *data, const ptrdiff_t *n, const ptrdiff_t *ni,                                 \
      const ptrdiff_t *local_ni, const ptrdiff_t *local_i_start,                        \
      MPI_Comm comm, int iterations, int iscale);                                       \
                                                                                        \
  PFFT_EXTERN int PX(check)(                                                            \
      const ptrdiff_t *n, MPI_Comm comm_cart_2d,                                        \
      unsigned pfft_flags, int verbose);                                                \
  PFFT_EXTERN int PX(check_many)(                                                       \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      const ptrdiff_t *block_col, const ptrdiff_t *block_row,                           \
      MPI_Comm comm_cart_2d, unsigned pfft_flags, int verbose);                         \
  PFFT_EXTERN int PX(check_guru)(                                                       \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
    const ptrdiff_t *block_col, const ptrdiff_t *block_row,                             \
    MPI_Comm comm_cart_2d, int verbose);                                                \
                                                                                        \
  PFFT_EXTERN ptrdiff_t PX(local_size_3d)(                                              \
      const ptrdiff_t *n, MPI_Comm comm_cart_2d, unsigned pfft_flags,                   \
      ptrdiff_t *local_ni, ptrdiff_t *local_i_start,                                    \
      ptrdiff_t *local_no, ptrdiff_t *local_o_start);                                   \
  PFFT_EXTERN ptrdiff_t PX(local_size_3d_many)(                                         \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      ptrdiff_t howmany, const ptrdiff_t *block_col, const ptrdiff_t *block_row,        \
      const ptrdiff_t *num_gcells_below, const ptrdiff_t *num_gcells_above,             \
      MPI_Comm comm_cart_2d, unsigned pfft_flags,                                       \
      ptrdiff_t *local_ni, ptrdiff_t *local_i_start,                                    \
      ptrdiff_t *local_no, ptrdiff_t *local_o_start,                                    \
      ptrdiff_t *alloc_local_data_in);                                                  \
  PFFT_EXTERN ptrdiff_t PX(local_size_3d_guru)(                                         \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
      ptrdiff_t howmany, const ptrdiff_t *block_col, const ptrdiff_t *block_row,        \
      const ptrdiff_t *num_gcells_below, const ptrdiff_t *num_gcells_above,             \
      MPI_Comm comm_cart_2d,                                                            \
      ptrdiff_t *local_ni, ptrdiff_t *local_i_start,                                    \
      ptrdiff_t *local_no, ptrdiff_t *local_o_start,                                    \
      ptrdiff_t *alloc_local_data_in, PX(permutation)* perm_out);                       \
                                                                                        \
  PFFT_EXTERN PX(plan) PX(plan_dft_3d)(                                                 \
      const ptrdiff_t *n, C *data_in, C *data_out,                                      \
      MPI_Comm comm_cart_2d,                                                            \
      int sign, unsigned pfft_flags, unsigned fftw_flags);                              \
  PFFT_EXTERN PX(plan) PX(plan_many_dft_3d)(                                            \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      ptrdiff_t howmany, const ptrdiff_t *block_col, const ptrdiff_t *block_row,        \
      const ptrdiff_t *num_gcells_below, const ptrdiff_t *num_gcells_above,             \
      C *data_in, C *data_out, MPI_Comm comm_cart_2d,                                   \
      int sign, unsigned pfft_flags, unsigned fftw_flags);                              \
  PFFT_EXTERN PX(plan) PX(plan_guru_dft_3d) (                                           \
      const ptrdiff_t *n, const ptrdiff_t *ni, const ptrdiff_t *no,                     \
      const PX(profile) *prf, const PX(permutation) *perm,                              \
      ptrdiff_t howmany, const ptrdiff_t *block_col, const ptrdiff_t *block_row,        \
      const ptrdiff_t *num_gcells_below, const ptrdiff_t *num_gcells_above,             \
      C *data_in, C *data_out, MPI_Comm comm_cart_2d,                                   \
      int sign, unsigned fftw_flags);                                                   \
                                                                                        \
  PFFT_EXTERN void PX(init_default_profile)(                                            \
      unsigned pfft_flags,                                                              \
      PX(profile) *prf, PX(permutation) *perm);                                         \
                                                                                        \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N1/np0 x N2  */                                \
  PFFT_EXTERN void PX(profile_n012_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  PFFT_EXTERN void PX(profile_n021_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */			       	\
  PFFT_EXTERN void PX(profile_n201_np012)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */                                \
  PFFT_EXTERN void PX(profile_n120_np012)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N1/np0 x N2/np1 x N0  */			        \
  PFFT_EXTERN void PX(profile_n120_np012_IN)(PX(profile) *prf);	                        \
  /*  n0/np0 x n1/np1 x n2  to  N2/np0 x N0/np1 x N1  */			        \
  PFFT_EXTERN void PX(profile_n201_np012_IN)(PX(profile) *prf);                         \
  PFFT_EXTERN void PX(profile_n201_np012_IN1)(PX(profile) *prf);                        \
  PFFT_EXTERN void PX(profile_n201_np012_IN2)(PX(profile) *prf);                        \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse  */                                                                  \
  PFFT_EXTERN void PX(profile_n210_np102)(PX(profile) *prf);                            \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse, out of place only  */                                               \
  PFFT_EXTERN void PX(profile_n021_np102_outofplace)(PX(profile) *prf);                 \
  /*  n0/np0 x n1/np1 x n2  to  N0/np1 x N2/np0 x N1  */                                \
  /*  self inverse  */                                                                  \
  PFFT_EXTERN void PX(profile_n021_np102_inplace)(PX(profile) *prf);                    \
                                                                                        \
  PFFT_EXTERN ptrdiff_t PX(prod_ptrdiff_t)(                                             \
      int d, const ptrdiff_t *vec);                                                     \
  PFFT_EXTERN int PX(equal_ptrdiff_t)(                                                  \
      int d, const ptrdiff_t *vec1, const ptrdiff_t *vec2);                             \
  PFFT_EXTERN void PX(vcopy_ptrdiff_t)(                                                 \
      int d, const ptrdiff_t *vec1,                                                     \
      ptrdiff_t *vec2);                                                                 \
  PFFT_EXTERN void PX(vadd_ptrdiff_t)(                                                  \
      int d, const ptrdiff_t *vec1, const ptrdiff_t *vec2,                              \
      ptrdiff_t *sum);                                                                  \
  PFFT_EXTERN void PX(vsub_ptrdiff_t)(                                                  \
      int d, const ptrdiff_t *vec1, const ptrdiff_t *vec2,                              \
      ptrdiff_t *sum);                                                                  \
                                                                                        \
  PFFT_EXTERN void *PX(malloc)(                                                         \
      size_t n, MPI_Comm comm);                                                         \
  PFFT_EXTERN void PX(free)(                                                            \
      void *p);                                                                         \
                                                                                        \
  PFFT_EXTERN void PX(get_arg_INT)(                                                     \
      int argc, char **argv, char *name, int neededArgs, ptrdiff_t *parameter);         \
  PFFT_EXTERN void PX(get_arg_int)(                                                     \
      int argc, char **argv, char *name, int neededArgs, int* parameter);               \
                                                                                        \
  PFFT_EXTERN void PX(exchange_ghostcells)(                                             \
      PX(plan) ths);                                                                    \
  PFFT_EXTERN void PX(reduce_ghostcells)(                                               \
      PX(plan) ths);                                                                    \
                                                                                        \
  void PX(apr_complex_3d)(                                                              \
      fftw_complex *data, ptrdiff_t *local_n, ptrdiff_t *local_n_start,                 \
      char *name, MPI_Comm comm);                                                       \
  void PX(apr_complex_permuted_3d)(                                                     \
      fftw_complex *data, ptrdiff_t *local_n, ptrdiff_t *local_n_start,                 \
      int perm0, int perm1, int perm2, char *name, MPI_Comm comm);




#define PFFT_MANGLE_DOUBLE(name) PFFT_CONCAT(pfft_, name)
#define PFFT_MANGLE_FLOAT(name) PFFT_CONCAT(pfftf_, name)
#define PFFT_MANGLE_LONG_DOUBLE(name) PFFT_CONCAT(pfftl_, name)

PFFT_DEFINE_API(
    PFFT_MANGLE_DOUBLE, FFTW_MANGLE_DOUBLE,
    double, fftw_complex)
PFFT_DEFINE_API(
    PFFT_MANGLE_FLOAT, FFTW_MANGLE_FLOAT,
    float, fftwf_complex)
PFFT_DEFINE_API(
    PFFT_MANGLE_LONG_DOUBLE, FFTW_MANGLE_LONG_DOUBLE,
    long double, fftwl_complex)



#define PFFT_TRANSPOSE_NONE       (0U)
#define PFFT_TRANSPOSE_012        (0U)
#define PFFT_TRANSPOSE_102        (1U<< 0)
#define PFFT_TRANSPOSE_021        (1U<< 1)
#define PFFT_TRANSPOSE_201        (1U<< 2)
#define PFFT_TRANSPOSE_120        (1U<< 3)

#define PFFT_FORWARD              (0U)
#define PFFT_BACKWARD             (1U<< 4)
#define PFFT_TRANSPOSED           (1U<< 5)
#define PFFT_P3DFFT               (1U<< 6)

#define PFFT_OUTOFPLACE           (0U)
#define PFFT_INPLACE              (1U<< 7)

#define PFFT_DEFAULT_BLOCK        FFTW_MPI_DEFAULT_BLOCK

#define PFFT_DEFAULT_BLOCKS       NULL
#define PFFT_NO_GCELLS            NULL

#ifdef __cplusplus
  }  /* extern "C" */
#endif /* __cplusplus */

#endif /* __PFFT_H__ */
